#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "music.h"
#include "m_irq.h"
#include "adldata.h" /* for Period[] */

/* 
 Supported commandinfobytes currently include:
 
  Axx (set speed (ticks))
  Bxx
  Cxx
  Dx0 (volume slide up   ticks*x)
  D0y (volume slide down ticks*y)
  DxF (increase volume by y)
  DFy (decrease volume by y)
  DFF (maximizes volume)

  EFx (freqdec)    (finally correct!! 2000-09-21 / 3.0.2)
  EEx (UNTESTED)   (extra fine slide down)
  Exx (freqslide)  (-"-)

  FFx (freqinc)    (-"-)
  FEx (UNTESTED)   (extra fine slide up)
  Fxx (freqslide)  (-"-)

  Hxx (nearly correct 2000-09-21 - at least it works on adlib)
  Kxx (-"-)

  Gxx (Should work on adlib. *UNTESTED*. / 3.0.2)
  Lxx (-"-)

  Jxx (arpeggio)
  Pxx (NON-STANDARD! Specify arpeggio speed - default=32h (50Hz))
      (If bit 7 set, then arpeggio is only with two notes, not three.)

  Xxx (panning, not fully supported for FM. 00..FF, 80h=middle)

  Qxy (under development - seems not to be working correctly.)

  SBx (patternloop, x>0?count:start)
  SCx (notecut in   x ticks)
  SDx (notedelay    x ticks)
  SEx (patterndelay x lines)

  MBx (NON-STANDARD: reverb)
  MCx (NON-STANDARD: chorus)

  Txx (set tempo)

  By definition, rows played in minute is 24 * tempo / ticks.
  That means, in a second, it is 2/5*tempo/ticks.
  Example: T7D, A06: tick length is 5 / (125+125) = 1 / 50 seconds,
                     six ticks in row, means row is 6 / 50 seconds,
                     25/3 = 8.333.. rows per second.
  Timer ticks per row = rate * speed * 5 / tempo / 2

  Exx and Fxx:

    For Exx, PeriodSlide =  xx
    For Fxx, PeriodSlide = -xx

    In each frame (there are Axx frames in a row)
      the Period of note will be increased by PeriodSlide.

  Hxy and Kxy:
  
    For Kxy: first do Dxy and then set xy to be the last Hxy xy.

    Vibrato amplitude: y/16       semitones
    Vibrato speed:     x*ticks/64 cycles in second
    Ticks is the value set by Axx -command.
  
  Gxx and Lxy:
  
    For Lxy: first do Dxy and then set xx to be the last Gxx xx.
    
    Tune the period by xx in each tick, until the period is correct.

  To make:

    Rxy
    Uxx

  To fix:

    SEx (does not appear to handle command repeatings correctly)
    SBx (this has broken in some way)

*/

volatile struct MusData MusData;

static int UnShiftVol(long v)
{
    return v / (1 << volshift);
}

static long ShiftVol(int v)
{
    return v * (1 << volshift);
}

static void M_NoteOff(int ch)
{
    MusData.driver->noteoff(ch);
}

static void M_MidiCtrl(int ch, int ctrl, int value)
{
    MusData.driver->midictrl(ch, ctrl, value);
}

static void M_Pan(int ch, int value)
{
    /* Pan limits: -128..+127 */
    MusData.driver->setpan(ch, value - 128);
}

static void M_PeriodTouch(int ch, unsigned Period)
{
    MusData.driver->hztouch(ch, Period ? PERIODCONST / Period : 0);
}

static void M_GmBend(int ch, int val)
{
    MusData.driver->bend(ch, val);
}

static void M_NoteOn(int ch, int note, long vol, unsigned Instru, unsigned Period)
{
    /* M_GmBend(ch, 0x2000);  * redundant */
    MusData.driver->patch(
      ch,
      MusData.Instr[Instru]->D,
      MusData.Instr[Instru]->GM,
      MusData.Instr[Instru]->Bank & 127);
    
    MusData.driver->noteon(
      ch,
      note,
      Period ? PERIODCONST / Period : 0,
      UnShiftVol(vol) * (MusData.Instr[Instru]->D[8] >> 2) / 63,
      UnShiftVol(vol));
}

static void M_Touch(int ch, unsigned Instru, long vol)
{
    MusData.driver->touch(
      ch,
      UnShiftVol(vol) * (MusData.Instr[Instru]->D[8] >> 2) / 63,
      UnShiftVol(vol));
}

void MusSeekRow(unsigned CutLine)
{
    MusData.posi.Pat = MusData.Orders[MusData.posi.Ord];
    MusData.posi.Pos = MusData.Patn[MusData.posi.Pat].Ptr;
    MusData.posi.Row = CutLine;
    MusData.posi.Wait = 0; /* I think this must be done */
    while(CutLine)
    {
        if(MusData.posi.Wait)
            --MusData.posi.Wait;
        else if(!MusData.posi.Wait)
        {
            BYTE ByteRead;
            while((ByteRead = *MusData.posi.Pos++) >= 16)
            {
                BYTE ByteRead = *MusData.posi.Pos++;
                if(ByteRead & 16)MusData.posi.Pos++; /* info */
                if(ByteRead & 32)
                {
                    BYTE a = *MusData.posi.Pos++; /* note */
                    if((!(MusData.Hdr->OrdNum&256)) || (a&128))
                        MusData.posi.Pos++; /* instru */
                }
                if(ByteRead & 64)MusData.posi.Pos++; /* volume */
                if(ByteRead & 128)MusData.posi.Pos++; /* cmd */
            }
            /* Skip blank lines */
            if(ByteRead == 15)
                MusData.posi.Wait = *MusData.posi.Pos++ - 14;
            else
                MusData.posi.Wait = ByteRead;
        }
        CutLine--;
    }
    /* At this point, we are always on a new order.             */
    /* SB0 must always be assumed upon incoming to a new order. */
    MusData.PendingSB0 = 1;
}

static unsigned MakePeriod(unsigned Note, unsigned Oct, unsigned C4Spd)
{
    return ((8363 * 16 * Period[11-Note]) >> Oct) / C4Spd;
}

static void CheckLackingCommands(const BYTE *const GotCmd)
{
    unsigned ch;
    for(ch=0; ch<MAXCHN; ch++)
    {
        BYTE Cmd = GotCmd[ch];

        if(Cmd != 'D' && Cmd != 'K' && Cmd != 'L')MusData.VolSlide[ch] = 0;
        if(Cmd != 'E' && Cmd != 'F')MusData.PerSlide[ch] = 0;
        if(Cmd != 'G' && Cmd != 'L')MusData.Portamento[ch] = 0;
        if(Cmd != 'J')MusData.Arpeggio[ch]=0;
        if(Cmd != 'Q')MusData.Retrig[ch]  =0;
        if(Cmd != 'H' && Cmd != 'K')
        {
            MusData.VibDepth[ch] =
            MusData.VibSpd  [ch] =
            MusData.VibPos  [ch] = 0;
        }
    }
}

static void PlayLine(void)
{
    struct MusPlayerPos ToSave;
    int Restore;
    int Cut, CutLine;
    int Break, BreakOrder;
    int CurChn;
    BYTE GotCmd[MAXCHN];
    BYTE Cmd;

    if(MusData.PatternDelay)
    {
        unsigned a;
        /* Set empty row, because we don't want *
         * to retrig the notes while doing SEx. */
        for(a=0; a<MAXCHN; MusData.CurPuh[a++] = 255);
        --MusData.PatternDelay;
        return;
    }

    ToSave = MusData.posi;

    memset(&GotCmd, 0, sizeof GotCmd);
    
    Restore = Cut = Break = 0;

    if(MusData.PendingSB0)
    {
        /* Do saving at the pattern first line! */
        MusData.saved = ToSave;
        MusData.PendingSB0 = 0;
    }

    Cmd = 0;

    if(MusData.posi.Wait)
        --MusData.posi.Wait;
    else if(!MusData.posi.Wait)
    {
        BYTE ByteRead;
        
        /* Rivi loppuu tavuun, jonka nelj ylint bitti ovat 0. */
        
        while((ByteRead = *MusData.posi.Pos++) >= 16)
        {
            BYTE Instru, Info;
    
            CurChn = ByteRead&15; /* There are no hdr.channels anymore */
            Instru = MusData.CurInst[CurChn];
            Info = 0;
    
            if(ByteRead & 16)
                Info = *MusData.posi.Pos++;
    
            MusData.CurPuh[CurChn] = 255;
                
            if(ByteRead & 32) /* Note(+instru) specified? */
            {
                BYTE a = Instru;
                BYTE Puh = *MusData.posi.Pos++;
    
                if(!(MusData.Hdr->OrdNum&256))
                {
                    /* If instrument data has not been optimized, instrument
                     * number always follows the note number.
                     * The optimization can not be used if there
                     * exist notes where oct*16+note >= 128.
                     */
                    
                    a = *MusData.posi.Pos++;
                }
                else if(Puh&128)
                {
                    /* If the optimization is applied, the instrument
                     * number existence is determined by Note&128.
                     * OrdNum&256 flags this situation.
                     */
                    
                    if(Puh < 254)Puh &= 127;
                    a = *MusData.posi.Pos++;
                }
                if(a)
                {
                    MusData.CurInst[CurChn] = Instru = a;
                    MusData.CurVol[CurChn] = ShiftVol(MusData.Instr[Instru]->Volume & 63);
                }
                
                MusData.CurPuh[CurChn] = Puh;
                
                if(Puh != 255)
                {
                    MusData.NoteCut[CurChn] = 0;
                    MusData.NoteDelay[CurChn] = 1;
                
                    if(Puh != 254)
                    {
                        int r;
                        
                        BYTE Oct  = Puh>>4;
                        BYTE Note = Puh&15;
                        
                        MusData.NNum[CurChn] = (12*Oct) + Note;
                        MusData.NNum[CurChn] += MusData.Instr[Instru]->FT - 2;
                        /* MIDI finetuning. -2 is a hack. */
            
                        r = ((MusData.Instr[Instru]->D[9] >> 2) & 15);
            
                        if(r)
                        {
                            r = rand() % (r + 1);
                            MusData.NoteDelay[CurChn] = r+1;
                        }
                        
                        /* AdLib finetuning also? */            
                        if(MusData.Instr[Instru]->Volume & 128)
                        {
                            /* Recalculate these */
                            Note = MusData.NNum[CurChn] % 12;
                            Oct  = MusData.NNum[CurChn] / 12;
                        }
                        
                        MusData.NextPeriod[CurChn] =
                            MakePeriod(Note, Oct, MusData.Instr[Instru]->C4Spd);
                    }
                }
            }
    
            if(ByteRead & 64) /* Volume specified? */
            {
                MusData.CurVol[CurChn] = ShiftVol(*MusData.posi.Pos++);
                MusData.NoteDelay[CurChn] = 1;
            }
    
            Cmd = '-';
            if(ByteRead & 128) /* Cmd(+infobyte) specified? */
                GotCmd[CurChn] = Cmd = (*MusData.posi.Pos++) | 64;
    
            switch(Cmd)
            {
                case 'A':
                    MusData.FrameCount = Info;
                    break;
                case 'B':
                    Break = 1;
                    /* The order is in hexadecimal format */
                    BreakOrder = Info;
                    break;
                case 'C':
                    Cut = 1;
                    /* The line is in bcd format */
                    CutLine = (Info>>4)*10 + (Info&15);
                    break;
                case 'D':
                case 'K':
                    if((ByteRead & 64) && !MusData.CurVol[CurChn])
                    {
                        /* FIXME: Any clue why is this here?        */
                        /* Does it have something to do with midi?  */
                        MusData.CurVol[CurChn] = ShiftVol(1);
                    }
    
                    if(!Info)Info = MusData.DxxDefault[CurChn];
                    else MusData.DxxDefault[CurChn] = Info;
    
                    if(Info==0xFF)MusData.CurVol[CurChn] = ShiftVol(63);
                                                             /* DFF, maximize */
                    else if(Info > 0xF0)                     /* DFx           */
                    {
                        MusData.VolSlide[CurChn] = -(Info&0x0F);
                        MusData.FineVolSlide[CurChn] = 1;
                    }
                    else if((Info&0x0F)==0x0F && (Info&0xF0)) /* DxF */
                    {
                        MusData.VolSlide[CurChn] = Info>>4;
                        MusData.FineVolSlide[CurChn] = 1;
                    }
                    else if(!Info);                           /* D00 */
                    else if(!(Info&0xF0))                     /* D0x */
                    {
                        MusData.VolSlide[CurChn] = -(Info&0x0F);
                        MusData.FineVolSlide[CurChn] = 0;
                    }
                    else if(!(Info&0x0F))                     /* Dx0 */
                    {
                        MusData.VolSlide[CurChn] = Info>>4;
                        MusData.FineVolSlide[CurChn] = 0;
                    }
                    if(Cmd=='K') { Info=0; goto CmdH; }
                    if(Cmd=='L') { Info=0; goto CmdG; }
                    break;
                
                case 'E':
                case 'F':
                    MusData.FinePerSlide[CurChn] = 0;
                    if(!Info)Info = MusData.ExxDefault[CurChn];
                    else MusData.ExxDefault[CurChn] = Info;
                    if(Info >= 0xF0)MusData.FinePerSlide[CurChn]=1, Info&=15;
                    else if(Info >= 0xE0)MusData.FinePerSlide[CurChn]=2, Info&=15;
                    MusData.PerSlide[CurChn] = (Cmd=='F') ? -Info : Info;
                    break;
                
                case 'G':
CmdG:               if(!Info)Info = MusData.GxxDefault[CurChn];
                    else MusData.GxxDefault[CurChn] = Info;
                    MusData.Portamento[CurChn] = Info;
                    break;
                
                case 'H':
CmdH:               if(!Info)Info=MusData.HxxDefault[CurChn];
                    else MusData.HxxDefault[CurChn]=Info;
                    if(Info & 0x0F)MusData.VibDepth[CurChn] = Info & 0x0F;
                    if(Info & 0xF0)MusData.VibSpd  [CurChn] = Info & 0xF0;
                    break;
                
                case 'J':
                    if(Info)
                        MusData.Arpeggio[CurChn] = Info;
                    break;
    
                case 'M':
                    if(!Info)Info = MusData.MxxDefault[CurChn];
                    MusData.MxxDefault[CurChn] = Info;
                    
                    /* MBx is reverb.
                     * MCx is chorus.
                     * MAx is supported also, because I've accidentally
                     *        wrote some songs using it... //Bisqwit
                     */
                    
                    if((Info>>4)==0xB
                    || (Info>>4)==0xA)
                    {
                        /* Reverb is sticky. */
                        M_MidiCtrl(CurChn, 91, (Info&15)<<3);
                    }
                    else if((Info>>4)==0xC)
                    {
                        /* Chorus is sticky. */
                        M_MidiCtrl(CurChn, 93, (Info&15)<<3);
                    }
                    break;
                case 'N':
                    /* Bend is not sticky.
                     * It will automatically reset upon next note.
                     * FIXME: This should be done AFTER NoteDelay!
                     */
                    
                    /* Bend limits: 0x0000 - 0x3FFF
                     * 0x80 << 6 = 0x2000
                     * 0x00 << 6 = 0x0000
                     * 0xFF << 6 = 0x3FE0 or similar
                     */
                    M_GmBend(CurChn, (unsigned)Info << 6);
                    break;
                case 'P':
                    if(Info)
                        MusData.ArpegSpd = Info;
                    break;
                case 'Q':
                    if(Info)
                        MusData.Retrig[CurChn] = Info;
                    break;
                case 'S':
                {
                    BYTE a, b;
                 
                    if(!Info)Info = MusData.SxxDefault[CurChn]; /* Notice */
                    MusData.SxxDefault[CurChn] = Info;          /* Notice */
                    
                    a = Info >> 4;
                    b = Info & 15;
                    
                    switch(a)
                    {
                        case 0xB:
                            if(!b)
                            {
                                /* ToSave is what we had on the beginning of line */
                                MusData.saved = ToSave;
                            }
                            else
                            {
                                /* If there was no active loop going on */
                                if(MusData.LoopCount == 0)
                                {
                                    /* Then set it */
                                    MusData.LoopCount = b;
                                }
                                else
                                {
                                    /* Else decrease the ongoing counter */
                                    --MusData.LoopCount;
                                }
                                
                                /* If there is still something to loop */
                                if(MusData.LoopCount)
                                {
                                    /* Then we'll go back when we can */
                                    Restore = 1;
                                }
                                else
                                {
                                    /* Else we're done, and we must do SB0 at next line. */
                                    MusData.PendingSB0 = 1;
                                }
                            }
                            break;
                        case 0xC:
                            /* Mitkhn ideaa olisi komennossa SC0... */
                            MusData.NoteCut[CurChn] = b + 1;
                            break;
                        case 0xD:
                            /* Mitkhn ideaa olisi komennossa SD0... */
                            MusData.NoteDelay[CurChn] = b + 1;
                            break;
                        case 0xE:
                            /* Mitkhn ideaa olisi komennossa SE0... */
                            MusData.PatternDelay = b;
                            break;
                    }
                    break;
                }
                
                case 'T':
                    MusData.Tempo = Info;
                    break;
                
                case 'X':
                    /* Pan is sticky. */
                    M_Pan(CurChn, Info);
                    break;
            }
        }
        
        /* Rivin lopettavan tavun (jonka ylimmt 4 bitti ovat 0)
         * nelj alinta bitti kertovat, montako tyhj rivi tmn
         * jlkeen seuraa.
         *    0: 0
         * 1-14: 2-15
         *   15: seuraava tavu kertoo mrn
         */
        
        if(ByteRead == 15)
            MusData.posi.Wait = *MusData.posi.Pos++ - 14;
        else
            MusData.posi.Wait = ByteRead;
    }
    
    /* Puuttuvat komennot ksitelln tss vaiheessa. */
    CheckLackingCommands(GotCmd);
    
    if(Restore)
    {
        /* If got SBx _now_ */
        MusData.posi = MusData.saved;
        return;
    }
    
    if(!MusData.PatternDelay)
        ++MusData.posi.Row;

    if(Break)
    {
        MusData.posi.Ord = BreakOrder;
        
        /* Bxx, hyptn mrttyyn orderiin.
         * Jos oli mys Cxx, hyptn mrtylle riville.
         */
        
        MusSeekRow(Cut ? CutLine : 0);
    }
    else if(Cut)
    {
        /* Cxx, hyptn seuraavaan orderiin mrtylle riville. */
        /* ++ -orderit yms pitisi loistaa poissaolollaan, joten tm toimii */
        MusData.posi.Ord = (MusData.posi.Ord+1) % (MusData.Hdr->OrdNum & 255);
        
        MusSeekRow(CutLine);
    }
    else if(!MusData.posi.Wait
         && MusData.posi.Pos >= MusData.Patn[MusData.posi.Pat].Ptr
                              + MusData.Patn[MusData.posi.Pat].Len)
    {
        /* Jos patterni loppui kesken, mennn seuraavaan orderiin, riville 0. */
        
        MusData.posi.Ord = (MusData.posi.Ord+1) % (MusData.Hdr->OrdNum & 255);
        
        MusSeekRow(0);
    }
}

static void UpdateEffects(void)
{
    unsigned ch;
    
    /* This loop must be called as often as possible.
     * The bigger rate, the better results.
     * Rate-variable must be correct, and indicate
     * the amount of times in second this loop is
     * executed.
     */
    for(ch=0; ch<MAXCHN; ch++)
    {
        /**** Jxy ****/
        
        /* Arpeggio has its own timer. */
        if(MusData.Arpeggio[ch])
        {
            WORD ArpCounter = (((unsigned long)(MusData.ArpegSpd & 127)) << 16) / Rate;
            
            MusData.ArpegClk += ArpCounter;
            if(MusData.ArpegClk < ArpCounter) /* was overflow? */
            {
                int notemp, Note;
                BYTE Instru = MusData.CurInst[ch];
                
                switch((MusData.ArpegCnt[ch]++) % (3-(MusData.ArpegSpd>>7)))
                {
                    case 0:
                        Note = 0;
                        break;
                    case 1:
                        Note = MusData.Arpeggio[ch] & 15;
                        break;
                    default:
                        Note = MusData.Arpeggio[ch] >> 4;
                }

                Note += MusData.NNum[ch];
                
                notemp = Note;

                if(!(MusData.Instr[Instru]->Volume & 128))
                {
                    notemp -= MusData.Instr[Instru]->FT;
                    /* Remove finetuning effect for AdLib */
                }
                
            //  M_NoteOff(ch);
                M_NoteOn(ch, Note, MusData.CurVol[ch], Instru,
                         MakePeriod(notemp%12, notemp/12,
                                    MusData.Instr[Instru]->C4Spd));
            }
        }
        
        /**** Hxy and Kxy ****/
        
        /* Vibrato has also its own timer. */
        if(MusData.ST3Period[ch] && MusData.VibDepth[ch])
        {
            /* Hxy: 
                     Vibrato amplitude: y/16       semitones
                     Vibrato speed:     x*ticks/64 cycles in second
                     Ticks is the value set by Axx -command.
            */

            /* This table was picked up from Mikmod. */
            static const BYTE avibtab[128] =
            {   0,1,3,4,6,7,9,10,12,14,15,17,18,20,21,23,
                24,25,27,28,30,31,32,34,35,36,38,39,40,41,42,44,
                45,46,47,48,49,50,51,52,53,54,54,55,56,57,57,58,
                59,59,60,60,61,61,62,62,62,63,63,63,63,63,63,63,
                64,63,63,63,63,63,63,63,62,62,62,61,61,60,60,59,
                59,58,57,57,56,55,54,54,53,52,51,50,49,48,47,46,
                45,44,42,41,40,39,38,36,35,34,32,31,30,28,27,25,
                24,23,21,20,18,17,15,14,12,10,9,7,6,4,3,1
            };

            WORD VibCounter = (((unsigned long)(MusData.VibSpd[ch] * MusData.FrameCount * 4)) << 16) / Rate;
            
            MusData.VibClk[ch] += VibCounter;
            if(MusData.VibClk[ch] < VibCounter) /* was overflow? */
            {
                /* Part of this code was also from Mikmod. */
                int VibVal = avibtab[MusData.VibPos[ch] & 127];
                
                if(MusData.VibPos[ch] & 0x80)VibVal = -VibVal;
                
                VibVal *= MusData.VibDepth[ch]; /* Range: -63*15..63*15 */
                
                M_PeriodTouch(ch, MusData.ST3Period[ch] + (VibVal >> 3));
                
                /* Our brand new midi vibrato! :)              */
                /* This is not _exact_, but it is near enough. */
                
                M_GmBend(ch, 0x2000 + VibVal*8);
                
                ++MusData.VibPos[ch];
            }
        }
    }
}

static void FrameTick(void)
{
    unsigned ch;
    
    for(ch=0; ch<MAXCHN; ch++)
    {
        if(MusData.NoteCut[ch] > 0)
        {
            if(!--MusData.NoteCut[ch])
                M_NoteOff(ch);
        }

        if(MusData.NoteDelay[ch] > 0)
        {
            /* Vibrato - don't use */
            /* M_MidiCtrl(CurChn, 1, 0); */
            
            if(!--MusData.NoteDelay[ch])
                switch(MusData.CurPuh[ch])
                {
                    case 255:   /* Tyhj rivi */
                        M_Touch(ch, MusData.CurInst[ch],
                                    MusData.CurVol[ch]);
                        break;
                    case 254:   /* ^^ -rivi */
                        M_NoteOff(ch);
                        break;
                    default:    /* Rivi jolla nuotti */
                        M_NoteOff(ch);
                        
                        if(!MusData.Portamento[ch])
                            MusData.ST3Period[ch] = MusData.NextPeriod[ch];
                        
                        M_NoteOn(ch, MusData.NNum[ch],
                                     MusData.CurVol[ch],
                                     MusData.CurInst[ch],
                                     MusData.ST3Period[ch]);
                }
                /* FIXME:
                   When PatternDelay is used, NoteDelay
                   should retrigger the same note again
                   and again.
                    !MusData.PatternDelay
                */
        }
        
        /**** Qxy ****/
        
        /* Don't retrig on the first frame */            
        if(MusData.Retrig[ch] && MusData.Frame)
        {
            unsigned space = MusData.Retrig[ch] & 15;
            unsigned fade  = MusData.Retrig[ch] >> 4;
            
            if(!(MusData.Frame % space))
            {
                MusData.CurVol[ch] -= ShiftVol(fade);
                if(MusData.CurVol[ch] < 0)MusData.CurVol[ch] = 0;
                
                M_NoteOff(ch);
                M_NoteOn(ch, MusData.NNum[ch],
                             MusData.CurVol[ch],
                             MusData.CurInst[ch],
                             MusData.ST3Period[ch]);
            }
        }
        
        /**** Dxy and Kxy and Lxy ****/
        
        if(MusData.VolSlide[ch])
        {
            long Vol = MusData.CurVol[ch];
            
            if(!MusData.Frame || !MusData.FineVolSlide[ch])
                Vol += ShiftVol(MusData.VolSlide[ch]);
                
            if(Vol < 0)Vol = 0;
            if(Vol > ShiftVol(63))Vol = ShiftVol(63);

            MusData.CurVol[ch] = Vol;
            
            M_Touch(ch,
              MusData.CurInst[ch],
              MusData.CurVol[ch]);

            if(MusData.FineVolSlide[ch] && !MusData.PatternDelay)
                MusData.VolSlide[ch] = 0;
        }
        
        /**** Exy and Fxy ****/

        if(MusData.PerSlide[ch])
        {
            if(!MusData.Frame || !MusData.FinePerSlide[ch])
            {
                int tmp = MusData.PerSlide[ch]
                        * ((MusData.FinePerSlide[ch] == 2) ? 1 : 4);
                MusData.ST3Period[ch] += tmp;
                  
                M_PeriodTouch(ch, MusData.ST3Period[ch]);
            }
            
            /* M_GmBend(ch, 0x2000); FIXME - not implemented. */

            if(MusData.FinePerSlide[ch] && !MusData.PatternDelay)
                MusData.PerSlide[ch] = 0;
        }
        
        /**** Gxx and Lxy ****/

        if(MusData.Portamento[ch])
        {
            int change = MusData.Portamento[ch] * 4, done=0;
            if(MusData.ST3Period[ch] < MusData.NextPeriod[ch])
            {
                MusData.ST3Period[ch] += change;
                if(MusData.ST3Period[ch] >= MusData.NextPeriod[ch])done=1;
            }
            else
            {
                MusData.ST3Period[ch] -= change;
                if(MusData.ST3Period[ch] <= MusData.NextPeriod[ch])done=1;
            }
            if(done)
            {
                MusData.ST3Period[ch] = MusData.NextPeriod[ch];
                MusData.Portamento[ch] = 0;
            }
            
            M_PeriodTouch(ch, MusData.ST3Period[ch]);
            
            /* M_GmBend(ch, 0x2000); FIXME - not implemented. */
        }
    }
}

/* Handles one timer tick */
/* Called from m_irq.c    */
void TimerHandler(void)
{
    if(MusData.Paused || !MusData.driver)return;
    
    MusData.FrameClk += MusData.FrameSpd;
    if(MusData.FrameClk < MusData.FrameSpd)
    {
        if(++MusData.Frame >= MusData.FrameCount)
        {
            ++MusData.NextRow;            
            PlayLine();
            
            MusData.Frame      = 0;
            
            /* This is: ((Tempo * 2 / 5) << 16) / Rate
             * but with more precision
             */
            
            MusData.FrameSpd = ((unsigned long)MusData.Tempo << 17) / (5 * Rate);
        }
        
        /* Frame has ticked, let's do something */
        
        ++MusData.NextFrame;        
        FrameTick();
    }
        
    /* Next we do all the frame independent stuff.          */
    /* Currently includes arpeggio (Jxy) and vibrato (Hxy). */
    
    ++MusData.NextTick;
        
    UpdateEffects();
}

void StartS3M(const BYTE *s)
{
    BYTE a;

    StopS3M();

    /* Pointers are nice :) */
    MusData.Hdr = (InternalHdr *)__FILE__" v"VERSION"(C) 1992,2000 Bisqwit (bisqwit@iki.fi)";
    /* Please don't modify the text above. It would be nice to   *
     * see my name inside games made by people unknown to me :-) */

    MusData.Playing = s;
    if(s[1]>2)for(;*s++;); /* Skip possible song name */

    MusData.Hdr = (const InternalHdr *)s;
    s += sizeof(InternalHdr);
    if(MusData.Hdr->PatNum & 512)
        s++; /* ChanCount present */

    MusData.Orders = (const BYTE *)s;
    s += (BYTE)MusData.Hdr->OrdNum;
    
    for(a=0; a<MusData.Hdr->InsNum; a++)
    {
        InternalSample *tmp = (InternalSample *)s;
        
        if(MusData.Hdr->PatNum & 512)
            s += 1 + strlen(tmp->insname);
       
        s += sizeof(*tmp);
        
        MusData.Instr[a+1] = tmp;
    }

    /* This disables some bugs. */
    MusData.Instr[0] = MusData.Instr[1];

    ((InternalHdr *)MusData.Hdr)->PatNum &= ~512;

    for(a=0; a<MusData.Hdr->PatNum; a++)
    {
        MusData.Patn[a].Len = ((BYTE)s[0]) | (((BYTE)s[1]) << 8);
        MusData.Patn[a].Ptr = (s += 2);
        s += MusData.Patn[a].Len;
    }

    for(a=0; a<MAXCHN; a++)
    {
        MusData.CurVol[a] = 0;
        MusData.CurInst[a] = 0;
        MusData.NoteDelay[a] = -1;
        MusData.VolSlide[a] = 0;
        MusData.PerSlide[a] = 0;
        MusData.Portamento[a] = 0;
        MusData.Retrig[a] = 0;
        MusData.Arpeggio[a] = 0;
    }
    
    SetPause(1);

    MusSeekOrd(0);
    MusData.LinearMidiVol = (MusData.Hdr->InSpeed & 128) == 128;
    MusData.Tempo      = MusData.Hdr->InTempo;
    MusData.FrameCount = MusData.Hdr->InSpeed & 127;
    MusData.ArpegSpd     = 50;
    MusData.GlobalVolume = 64;
    
    SetPause(0);
}

/* From m_arch.c */
extern int FromArchive;
extern char *Archived;

void StopS3M(void)
{
    if(Archived && !FromArchive)
    {
        free(Archived);
        Archived = NULL;
    }
    
    SetPause(1);
    
    ReleaseINT();
}

void SetPause(int state)
{
    if(state == MusData.Paused)return;
    MusData.Paused = state;
    
    if(MusData.driver)
        MusData.driver->reset();

    if(state) /* pause */
    {
        /* actually nothing to do */
    }
    else      /* unpause */
    {
        InstallINT();
    }
}

void PauseS3M(void)
{
    if(MusData.Paused)
        SetPause(0);
    else
        SetPause(1);
}

void MusSeekOrd(unsigned Num)
{    
    int p = MusData.Paused;
    
    SetPause(1);
    
    /* The player must not modify these at same time */
    MusData.PatternDelay = 0;
    MusData.posi.Wait = 0;
    MusData.posi.Row = 0;
    MusData.posi.Ord = Num;
    MusData.posi.Pat = MusData.Orders[Num];
    MusData.posi.Pos = MusData.Patn[MusData.posi.Pat].Ptr;
    /* SB0 must always be done at the beginning of order */
    MusData.PendingSB0 = 1;
    MusData.LoopCount  = 0;
    MusData.FrameSpd   = 0xFFFF; /* Hurry first. */
    MusData.Frame      = MusData.FrameCount-1;
    
    SetPause(p);
}

void MusSeekNextOrd(void)
{
    if(MusData.posi.Ord < MusData.Hdr->OrdNum - 1)
        MusSeekOrd(MusData.posi.Ord+1);
}

void MusSeekPrevOrd(void)
{
    if(MusData.posi.Ord > 0)MusSeekOrd(MusData.posi.Ord - 1);
}
